// Gray code exploration program
// Michael Ernst <mernst@alum.mit.edu>
// August 16, 1999

// Goal: discover cyclic Gray codes (the last element differs from the
// first by only one bit) such that there is no length-2 cycle in the graph
// induced by the Gray code.

#include <iostream>
#include <string>
#include <strstream>
// To disable the assertions (and dramatically improve runtime),
// compile with the "-DNDEBUG" compiler flag.  (Don't forget -O in that case.)
#include <assert.h>
#ifndef NODMALLOC
#include "dmalloc.h"
#endif
#include "gray-codes.h"

// Example compilation lines:
//  For debugging:   g++ -Wall -g -o gray-codes gray-codes.cpp
//  For speed:       g++ -Wall -DNDEBUG -O3 -o gray-codes gray-codes.cpp


// Print all Gray codes that are at least this long.
// If this value were STATES, we would print only cyclic Gray codes.
// This is not STATES (ie, we print even non-cyclic Gray codes) because I
// want to see those for debugging and because a subsequent pass does
// additional filtering.
const int NUM_TRANS_FOR_PRINTING = STATES-1;


int main(int argc, char *argv[]) {
  cout << "Computing Gray codes for " << BITS << " bits.\n";
  
  bool just_one_cyclic = false;
  int progress_num_trans = 0;

  Gray_Code gc;
  Gray_Code final_gc;
  int final_num_trans = 0;

  // Process arguments
  {
    int curr_arg = 1;
    bool found_induced_constraint = false;
    while (curr_arg < argc) {
      if (strcmp("-initial", argv[curr_arg]) == 0) {
	if (curr_arg + 1 >= argc) {
	  cerr << "-initial flag requires an argument\n";
	  exit(-1);
	}
	gc.add_transitions(argv[curr_arg+1]);
	curr_arg += 2;
      } else if (strcmp("-final", argv[curr_arg]) == 0) {
	if (curr_arg + 1 >= argc) {
	  cerr << "-final flag requires an argument\n";
	  exit(-1);
	}
	final_gc.add_transitions(argv[curr_arg+1]);
	curr_arg += 2;
      } else if (strcmp("-edgefile", argv[curr_arg]) == 0) {
	if (graph_constraint != obey_input_graph) {
	  cerr << "Not compiled with support for -edgefile argument.\n";
	  exit(-1);
	}
	if (curr_arg + 1 >= argc) {
	  cerr << "-edgefile flag requires an argument\n";
	  exit(-1);
	}
	if (found_induced_constraint) {
	  cerr << "Multiple -edgefile arguments supplied\n";
	  exit(-1);
	}
	int input_induced_contraint[BITS][BITS];
	Gray_Code::read_square_boolean_array(argv[curr_arg+1], input_induced_contraint);
	curr_arg += 2;
	gc.set_induced(input_induced_contraint);
	found_induced_constraint = true;
      } else if (strcmp("-justonecyclic", argv[curr_arg]) == 0) {
        just_one_cyclic = true;
	curr_arg++;
      } else if (strcmp("-progress", argv[curr_arg]) == 0) {
	if (curr_arg + 1 >= argc) {
	  cerr << "-progress flag requires an argument\n";
	  exit(-1);
	}
	progress_num_trans = str2int(argv[curr_arg+1]);
        if ((progress_num_trans <= 0) || (progress_num_trans >= STATES)) {
          cerr << "-progress argument should be between 0 and " << STATES
               << "; got " << progress_num_trans << "\n";
          abort();
        }
	curr_arg += 2;
      } else {
	cerr << "Bad argument " << argv[curr_arg] << "\n";
	exit(-1);
      }
    }
    if ((graph_constraint == obey_input_graph) && !found_induced_constraint) {
      cerr << "No graph constraint (-edgefile argument) found.\n";
      exit(-1);
    }
  }
  
  // Iterate through all possible Gray codes.
  // There are n possible bit positions at each of 2^n sequence positions.
  // Stopping condition:  no more Gray codes to examine.


  // Given a candidate bit position:
  //  * if it is legal, go to the next state and choose candidate bit position 0.
  //  * if it is not legal, try to increment the candidate bit position.
  //     * if we succeed, then repeat with that candidate bit position.
  //     * if we fail to increment, go to the previous state and increment its bit position.
  //       We fail to increment (we are done with this state) if:
  //       [is this "or" or "and"?]
  //        * num_changes[prev_bit] == 0
  //        * we have examined all n possible bit positions

  // maximum number of transitions seen so far
  int num_trans_max = 0;

  bit_pos candidate_bit_pos = 0;


  while (gc.num_trans > -1) {
    assert(gc.check_invariants());

    state candidate_state = gc.new_state(candidate_bit_pos);

	// cout << "trying (len " << gc.num_trans << "): ";
	// for (int i=0; i<gc.num_trans; ++i) cout << gc.transition_bit[i] << " ";
	// cout << candidate_bit_pos;
	// cout << "\n";

    if ((gc.num_trans < STATES)
        && gc.legal_transition(candidate_state, candidate_bit_pos)) {
      gc.add_transition(candidate_state, candidate_bit_pos);

      // // This test can be satisfied, but I want to do those tests elsewhere
      // // (so as not to clutter the output of this program).
      // if (num_trans == STATES) {
      //   cout << "===========================================================================\n";
      //   if (candidate_state == 0)
      //     cout << "Yay!  Cyclic complete Gray code!\n";
      //   else
      //     cout << "Non-cyclic complete Gray code.\n";
      //   print_gray_code();
      // } 
      // // else if (num_trans == STATES-1) {
      // //   cout << "Near miss:\n";
      // //   print_gray_code();
      // // }

      candidate_bit_pos = 0;

      if (gc.num_trans <= progress_num_trans) {
        cout << "Progress:";
        for (int i=0; i<gc.num_trans; ++i) {
          cout << " " << gc.transition_bit[i];
        }
        cout << "\n";
        cout.flush();
      }

    } else {
      // Try to increment the candidate Gray code
      
      while ((candidate_bit_pos == BITS-1)
	     // Do not skip over a bit position with precedence over it
	     // but for which we have not yet taken a transition.
	     || ((graph_constraint == no_bidirectional_edges)
		 && (gc.num_changes[candidate_bit_pos] == 0))) {
	// We can't increment this bit, so pop the stack.
	// If this is the longest Gray code found so far, or if this is a
	// nearly complete (or potentially fully complete) Gray code, then
	// print it as well.

        // Do not print the Gray code if it contains all transitions (and
        // returns back to the original state) but the first transition
        // isn't legal after the closing one.
        if (!((gc.num_trans == STATES)
              && (! gc.legal_transition(gc.transition_state[0],
                                        gc.transition_bit[0])))) {
          // First test is >= so we can see multiple instances of
          // the maximum length.
          if ((gc.num_trans >= num_trans_max)
              || (gc.num_trans >= NUM_TRANS_FOR_PRINTING)) {
            if (gc.num_trans > num_trans_max)
              num_trans_max = gc.num_trans;
            // gc.print_gray_code();
            gc.print_gray_code_short();
          }
        }
	
	// cout << "backing off (len " << gc.num_trans << "): ";
	// for (int i=0; i<gc.num_trans; ++i) cout << gc.transition_bit[i] << " ";
	// cout << candidate_bit_pos;
	// cout << "\n";
	
        if (just_one_cyclic && (gc.num_trans == STATES)) {
          cout << "Found a cyclic code; quitting.\n";
          exit(0);
        }

	if (gc.num_trans == 0) {
	  cout << "All done.\n";
	  exit(0);
	} else if (gc.num_trans == final_num_trans) {
          cout << "Should check for final\n";
        }
	
	// This call sets candidate_bit_pos and candidate_state
	gc.remove_transition(candidate_state, candidate_bit_pos);
	
        // if (gc.num_trans <= progress_num_trans) {
        //   cout << "backed off (len " << gc.num_trans << "):";
        //   for (int i=0; i<gc.num_trans; ++i)
        //     cout << " " << gc.transition_bit[i];
        //   cout << ", " << candidate_bit_pos;
        //   cout << "\n";
        // }
      }
      ++candidate_bit_pos;
      // Re-enable this once "orbit" is re-introduced and kept up to date.
      // if ((graph_constraint == obey_input_graph)
      //     && (gc.num_trans == 0)) {
      //   while ((candidate_bit_pos < BITS)
      //          && (orbit[candidate_bit_pos] != candidate_bit_pos)) {
      //     cout << "Skipping codes with first bit position " << candidate_bit_pos
      //          << " because isomorphic to " << orbit[candidate_bit_pos] << "\n";
      //     candidate_bit_pos++;
      //   }
      //   if (candidate_bit_pos == BITS) {
      //     cout << "All done.\n";
      //     exit(0);
      //   }
      // }       

    }
  }
}
 
