// gray-codes.h		-*- C++ -*-
// Gray codes data structure
// Michael Ernst <mernst@alum.mit.edu>
// August 23, 1999

#include <fstream>
#include <cstdlib>

// This class supports creating and searching for Gray codes, with an
// emphasis on their induced graphs.

// A Gray code is a sequence of length-n binary numbers such that two
// adjacent elements in the sequence differ in only one bit.
// In a circular Gray code, the last element also differs from the first in
// only one bit.
// I will number the bits 0 to n-1, per computer science convention.

// The induced graph is created in the following way:
//   The graph has n nodes.
//   There is a (directed) edge from node i to node j if at some point in
//   the Gray code, immediately after bit i has changed, the next bit to
//   change is j.

// This code supports two ways to constrain (the induced graph of) Gray codes:
//  * The user can supply an induced graph.
//    We don't need to start our search at every node, only at one from
//    each group of always-isomorphic nodes.  That isn't implemented yet, though.
//  * The user can supply constraints over the induced graph.
//    Without loss of generality, let the bits first change in order; that
//    is, the first bit to change is bit 0; the next bit after that to
//    change is bit 1; the next bit after that to change is bit 2, etc.
//    One plausible sequence of bit changes is 0,1,2,3,4,; another one is
//    0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,.  A special case of this is, "Without
//    loss of generality, make the first bit to change be bit 0."

// To do:

//  * When using no-bi-directional constraint, we require the first bit
//    changed to be 0.  We must not do that when the user supplies an
//    induced graph; fix this!
//  * When the user supplies an induced graph as a constraint, compute its
//    symmetries (its weak self-isomorphisms) and start at only one of the
//    edges in each such group.

//  * [I needn't worry about this, because it can't happen:  we have to
//     transition to a bit as often as we transition away from it.  There's
//     perhaps something related with some bits becoming inaccessible, but
//     that seems hard to test and the search will quit after at most k*2^k
//     state explorations, where k is the number of remaining accessible
//     bits.]  Perhaps add the following pruning heuristic:
//     If at some point a bit cannot be transitioned to (because it
//     transitions to every other bit), or it cannot be transitioned from
//     (because every other bit transitions to it), then cut off the search
//     immediately (actually, only if there exist unreached states with a
//     different value of the bit than its current value; but that's
//     guaranteed) rather than doing the exhaustive deeper search to
//     discover failure.  That seems an expensive test, but the odds of
//     things being OK when


// BITS, STATES, and graph_constraint are constants rather than arguments
// to the program because I want to squeeze out every last iota of speed.
// When compiling this program, you should probably define the preprocessor
// macros MBITS and MCONSTRAINT (both small integers).
// I use const variables rather than Cpp macros for better error checking;
// I'm hoping the compiler will do effective constant folding.

#ifndef MBITS
#define MBITS 6
#endif
const int BITS = MBITS;
const int STATES = 1 << BITS;	// also the maximum number of transitions

// How do we constrain the set of Gray codes we are examining?
// Right now there are the following possibilities:
//  * no constraint
//  * prohibit bidirectional edges
//  * permit only edges from a given graph (input by the user)
enum Graph_constraint { no_constraint=0, no_bidirectional_edges, obey_input_graph };
#ifndef MCONSTRAINT
#define MCONSTRAINT no_bidirectional_edges
#endif
const Graph_constraint graph_constraint = MCONSTRAINT;
const bool maintain_induced = (graph_constraint == no_bidirectional_edges);

typedef int state;
typedef int bit_pos;
typedef int seq_no;
typedef int state_count;


class Gray_Code {

  // everything is public, for convenience
 public:

  ///
  /// Data structures
  ///

  // A stack of (up to) (2^n)-1 bit positions (elements from 0 to n-1).
  // These represent the bit being changed at each transition.
  // It is indexed by sequence numbers, not by states.
  // Perhaps use a real stack?
  bit_pos transition_bit[STATES];
  // This is a separate variable for convenience.  Should it be?
  // Perhaps better to use an inline function to retrieve from stack.
  bit_pos prev_bit_pos;		// the top element of the stack

  // A stack of the states.  I could recompute, but this is convenient.
  // The very first state, 0, doesn't appear; these are the states after
  // the corresponding transition (in transition_bit[]).
  state transition_state[STATES];
  // This is a separate variable for convenience.  Should it be?
  // Perhaps better to use an inline function to retrieve from stack.
  state prev_state;		// the top element of the stack

  // [I call a bit change a transition -- from one state to the next.  The
  // paper uses "transition" for a transition from one bit change to the
  // next -- an edge in the induced graph.]
  // This is the number of transitions (bit changes).  The last transition
  // is at index one less than this.
  // The length of the Gray code is one more than the number of transitions.
  seq_no num_trans;


  // The induced graph, as a n-by-n matrix of integers.  Each integer
  // represents the number of crossings of that edge, so that we can
  // easily increment, decrement, and check for zero.
  // In other words, induced[i][j] > 0 implies we changed bit i, then
  // changed bit j, in the Gray code.
  // If maintain_induced is true:  the code maintains/updates this graph.
  //     In this case, we never add the circular edge to the induced graph.
  // If graph_constraint is obey_input_graph, this graph is supplied by the
  //     user.  In this case, it should include the circular edge.
  // In order to avoid confusion (and avoid overwriting the input graph),
  // we use a separate graph when checking for isomorphism.
  int induced[BITS][BITS];

  ///
  /// Constraining which Gray codes are explored.
  ///

  /// For all types of constraints.

  // A lookup table of which codewords positions have been reached.  This is
  // just an array indexed by integers (states).
  bool reached[STATES];


  ////////////////
  // Only for (graph_constraint == obey_input_graph)
  //
  
  void set_induced(int induced_constraint[BITS][BITS]) {
    assert(graph_constraint == obey_input_graph);
    for (int i=0; i<BITS; ++i)
      for (int j=0; j<BITS; ++j)
	induced[i][j] = induced_constraint[i][j];
  }


  // 
  // End of variables for (graph_constraint == obey_input_graph).
  ////////////////


  ////////////////
  // The following variables are used only when
  // (graph_constraint == no_bidirectional_edges).
  // 

  // The number of times a particular bit has been changed.  This is used
  // for exploiting ordering-of-bit-changes symmetry group.  It's an
  // integer rather than a boolean so that we can easily increment and
  // decrement.
  int num_changes[BITS];
  
  /// I think this may be irrelevant now that I'm using a separate graph for
  /// testing isomorphism.
  // // True if we have added to the count for the induced edge from the end to the
  // // beginning.  Should be set only when we won't be changing the code any
  // // more.
  // bool induced_circular;	       
  // void set_induced_circular() {
  //   assert(num_trans == STATES);
  //   bit_pos first_bit = transition_bit[0];
  //   bit_pos final_bit = transition_bit[STATES-1];
  //   // Possibly I should check that this is actually legal before doing
  //   // it. But I probably already did that; believe it?
  //   if (!induced_circular) {
  //     induced[final_bit][first_bit]++;
  //     induced_circular = true;
  //   }
  // }

  // 
  // End of variables for (graph_constraint == no_bidirectional_edges).
  ////////////////



  ///
  /// Initializers (Constructors)
  ///

  Gray_Code() {
    initialize();
  }

  Gray_Code(const char * s) {
    initialize();
    add_transitions(s);
  }

  Gray_Code(string s) {
    initialize();
    add_transitions(s.c_str());
  }

  void initialize() {
    prev_bit_pos = BITS;
    prev_state = 0;
    num_trans = 0;

    reached[0] = true;
    for (int i=1; i<STATES; ++i) {
      reached[i] = false;
    }
    
    graph = NULL;

    if (maintain_induced) {
      for (int i=0; i<BITS; ++i) {
	num_changes[i] = 0;
	for (int j=0; j<BITS; ++j) {
	  induced[i][j] = 0;
	}
      }
      // induced_circular = false;
    }

    switch (graph_constraint) {
    case no_constraint:
      break;
    case no_bidirectional_edges:
      break;
    case obey_input_graph:
      break;
    default:
      assert(0);
      break;
    }

    check_invariants();
    
  }

  ///
  /// Destructor
  ///

  ~Gray_Code() {
    delete graph;
  }

  ///
  /// Invariants
  ///

  bool check_invariants() {
    // cout << "starting invariant check\n";

    assert(reached[0] == true);
    for (int i=0; i<num_trans; ++i)
      assert(reached[transition_state[i]]);
    {
      int num_reached = 0;
      for (int i=0; i<STATES; ++i)
        if (reached[i])
          num_reached++;
      assert(((num_reached == STATES) && (num_trans == STATES))
             || (num_reached == num_trans+1));
    }

    assert(prev_bit_pos == ((num_trans == 0) ? BITS : transition_bit[num_trans-1]));
    assert(prev_state == ((num_trans == 0) ? 0 : transition_state[num_trans-1]));

    if (maintain_induced) {
      // // This should take account of induced_circular, shouldn't it?

      int induced_check[BITS][BITS];
      for (int i=0; i<BITS; ++i)
	for (int j=0; j<BITS; ++j)
	  induced_check[i][j] = 0;
      for (int i=1; i<num_trans; ++i)
	induced_check[transition_bit[i-1]][transition_bit[i]]++;
      for (int i=0; i<BITS; ++i)
	for (int j=0; j<BITS; ++j)
	  assert(induced_check[i][j] == induced[i][j]);
    }

    switch (graph_constraint) {
    case no_constraint:
      break;
    case no_bidirectional_edges:
      for (int i=0; i<BITS; ++i) {
	{
	  int sum1 = 0;
	  int sum2 = 0;
	  for (int j=0; j<BITS; ++j) {
	    sum1 += induced[i][j];
	    sum2 += induced[j][i];
            assert((induced[i][j] == 0) || (induced[j][i] == 0));
	  }
	  // I need to add an extra 1 for the last bit change, which doesn't
	  // have an induced "from" edge.  (Also for the first bit change,
	  // which doesn't have an induced "to" edge.)
	  if (num_trans > 0) {
	    if (i == transition_bit[0]) sum2++;
	    if (i == transition_bit[num_trans-1]) sum1++;
	  }

	  assert(num_changes[i] == sum1);
	  assert(num_changes[i] == sum2);
	}
	{
	  int count = 0;
	  for (int j=0; j<num_trans; j++) {
	    if (i == transition_bit[j]) {
	      ++count;
	    }
	  }
	  assert(num_changes[i] == count);
	}
      }
      break;
    case obey_input_graph:
      // Ought to check that the edges all obey induced[][].
      break;
    default:
      assert(0);
      break;
    }

    // cout << "invariant check succeeded\n";
    return true;
  }


  ///
  /// Operators
  ///

  // class version of new_state
  inline static state new_state(state old_state, bit_pos bit) {
    assert((bit >= 0) && (bit < BITS));
    return old_state ^ (1 << bit);
  }
  
  // member version of new_state
  inline state new_state(bit_pos bit) {
    return Gray_Code::new_state(prev_state, bit);
  }

  inline bool legal_transition(state candidate_state, bit_pos bit) {

    assert((bit >= 0) && (bit < BITS));

    // Any first transition is legal.
    if (num_trans == 0)
      return true;

    bool closing_trans = ((num_trans == STATES-1)
                          && (candidate_state == 0));
    // We can determine whether a full cyclic Gray code is legal by doing
    // legal_transition(gc.transition_state[0], gc.transition_bit[0]).
    bool post_closing_trans = ((num_trans == STATES)
                               && (candidate_state == transition_state[0]));

    // The state it transitions to has been reached already
    // (unless this is the closing bit change).
    if (reached[candidate_state]
        && (! closing_trans)
        && (! post_closing_trans))
      return false;

    if (graph_constraint == no_bidirectional_edges) {
      // This induces a 2-cycle in the induced graph, because we previously
      // transitioned using this bit before using whatever bit was just used.
      if (induced[bit][prev_bit_pos] != 0)
        return false;
    }      

    if (graph_constraint == obey_input_graph) {
      // Transition is not permited in the user-supplied graph.
      if (! induced[prev_bit_pos][bit])
        return false;
    }

    return true;

  }


  inline void add_transition(state candidate_state, bit_pos candidate_bit_pos) {
    assert((candidate_bit_pos >= 0) && (candidate_bit_pos < BITS));
    assert(candidate_state == new_state(candidate_bit_pos));
    assert(num_trans < STATES); // required lest arrays overflow
    assert(legal_transition(candidate_state, candidate_bit_pos));
    assert(graph == NULL);
    transition_bit[num_trans] = candidate_bit_pos;
    transition_state[num_trans] = candidate_state;
    ++num_trans;
    
    reached[candidate_state] = true;

    if (maintain_induced) {
      if (num_trans > 1) {
	induced[prev_bit_pos][candidate_bit_pos]++;
      }
    }

    if (graph_constraint == no_bidirectional_edges) {
      num_changes[candidate_bit_pos]++;
    }
    
    prev_bit_pos = candidate_bit_pos;
    prev_state = candidate_state;
    
    assert(check_invariants());
  }
  
  inline void remove_transition() {
    state dummy_removed_state;
    bit_pos dummy_removed_bit_pos;
    remove_transition(dummy_removed_state, dummy_removed_bit_pos);
  }

  // These are out-parameters:  they are not read, only set, by this routine.
  inline void remove_transition(state & removed_state, bit_pos & removed_bit_pos) {
    assert(num_trans > 0);
    assert((num_trans == STATES) == (prev_state == 0));

    removed_state = prev_state;
    removed_bit_pos = prev_bit_pos;
    --num_trans;
    if (num_trans > 0) {
      prev_bit_pos = transition_bit[num_trans-1];
      prev_state = transition_state[num_trans-1];
    } else {
      assert(num_trans == 0);
      prev_bit_pos = BITS;
      prev_state = 0;
    }
    
    if (removed_state != 0)
      reached[removed_state] = false;

    if (maintain_induced) {
      if (num_trans > 0) {
	induced[prev_bit_pos][removed_bit_pos]--;
      }
    }
    if (graph_constraint == no_bidirectional_edges) {
      num_changes[removed_bit_pos]--;
    }

    assert(check_invariants());
  }
  
  void add_transitions(const char * s) {
    assert(check_invariants());
    
    istrstream is(s);
    bit_pos candidate_bit_pos;
    while (is >> candidate_bit_pos) {
      assert(check_invariants());
      state candidate_state = new_state(candidate_bit_pos);
      
      assert(legal_transition(candidate_state, candidate_bit_pos));
      add_transition(candidate_state, candidate_bit_pos);
    }
    
    assert(check_invariants());
  }

  ///
  /// Graph isomorphism
  ///

  // The purpose of this is to assist in checking for isomorphism.
  // Isomorphism checking isn't a time-critical operation.
  class Graph {
    int induced[BITS][BITS];
    int indegree[BITS];
    int outdegree[BITS];
    int num_edges;

  public:
    Graph(Gray_Code & gc, bool cyclic) {
      for (bit_pos i=0; i<BITS; ++i) {
	for (bit_pos j=0; j<BITS; ++j)
	  induced[i][j] = 0;
	indegree[i] = 0;
	outdegree[i] = 0;
      }
      for (state i=1; i<gc.num_trans; i++)
	induced[gc.transition_bit[i-1]][gc.transition_bit[i]]++;
      if (cyclic)
	induced[gc.transition_bit[gc.num_trans-1]][gc.transition_bit[0]]++;
      for (bit_pos i=0; i<BITS; ++i)
	for (bit_pos j=0; j<BITS; ++j)
	  if (induced[i][j] > 0) {
	    outdegree[i]++;
	    indegree[j]++;
	  }
      num_edges = 0;
      for (int i=0; i<BITS; ++i)
	for (int j=0; j<BITS; ++j)
	  if (induced[i][j] != 0)
	    ++num_edges;
    }

    int edges() {
      return num_edges;
    }

    void print_induced() {
      cout << "Induced graph edges (" << num_edges << " total):\n";
      print_square_boolean_array(induced);
    }

    // candidate[BITS][BITS] indicates whether two bit positions are
    // permitted to potentially map to one another.  For now we use a
    // simple test that requires the indegrees and outdegrees to match.
    void set_candidate(Graph & other, bool candidate[BITS][BITS]) {
      for (bit_pos i=0; i<BITS; ++i)
	for (bit_pos j=0; j<BITS; ++j)
	  candidate[i][j] = ((indegree[i] == other.indegree[j])
			     && (outdegree[i] == other.outdegree[j]));
    }

    // Isomorphism only up to presence/absence of edges
    bool isomorphic_weak(Graph & other, bit_pos map[BITS]) {
      assert(valid_map(map));

      // cout << "Checking for (weak) edge isomorphism:\n";
      // print_map(map);
      // print_induced();
      // other.print_induced();
      
      for (bit_pos i=0; i<BITS; ++i)
	for (bit_pos j=0; j<BITS; ++j)
	  if ((induced[i][j] > 0) != (other.induced[map[i]][map[j]] > 0))
	    return false;
      
      return true;
    } 

    // Isomorphism up to number of crossings of each edge
    bool isomorphic_strong(Graph & other, bit_pos map[BITS]) {
      assert(valid_map(map));

      // cout << "Checking for strong edge isomorphism:\n";
      // print_map(map);
      // print_induced();
      // other.print_induced();
      
      for (bit_pos i=0; i<BITS; i++)
	for (bit_pos j=0; j<BITS; j++)
	  if (induced[i][j] != other.induced[map[i]][map[j]])
	    return false;
      
      return true;
    } 

  };

  // This is non-null if we are done adding to the Gray code and ready to
  // check isomorphism (of the Gray codes themselves or just of their
  // induced graphs).
  Graph *graph;
  // This could also be called prepare_for_isomorphism_testing.
  void set_graph(bool cyclic) {
    if (graph == NULL)
      graph = new Graph(*this, cyclic);
  }
    
  void set_graph() {
    bool cyclic = (num_trans == STATES);
    set_graph(cyclic);
  }

  // Add all_isomorphisms() iterator, perhaps.
  // Or maybe I should find someone else's code for doing graph
  // isomorphisms.  Mine is specialized in some ways, but still...

  // Return true iff the induced graphs are isomorphic.
  // Ignores the actual sequence of bit changes.
  bool isomorphic(Gray_Code & other, bool weak = false) {

    set_graph();
    other.set_graph();
    if (graph -> edges() != other.graph -> edges())
      return false;

    // Indicates whether two bit positions are permitted to potentially map
    // to one another.
    bool candidate[BITS][BITS];
    graph -> set_candidate(*(other.graph), candidate);
    // cout << "Isomorphism candidates:\n";
    // // Does candidate need to be an int array in order for this to work?
    // print_square_boolean_array(candidate);
    
    // Iterate through all the possible graph isomorphisms.
    // Similar to code in gray-codes.cpp.  Perhaps generalize to an iterator?

    bit_pos map[BITS];
    // inv_map[target] indicates whether, for some source, map[source] == target
    bool inv_map[BITS];
    for (bit_pos i=0; i<BITS; ++i) {
      map[i] = -1;
      inv_map[i] = false;
    }

    // Treat map like a stack when we iterate through all possible candidate
    // isomorphisms.
    bit_pos map_elts = 0;

    // The element we are considering adding to the stack
    bit_pos candidate_iso = 0;

    while (map_elts > -1) {
      // Check invariants
      assert(valid_inv_map(inv_map, map, map_elts));
      assert(candidate_iso < BITS);

      bool must_increment = false;

      // cout << "trying (len " << map_elts << "):";
      // for (int i=0; i<map_elts; ++i) cout << " " << map[i];
      // cout << ", " << candidate_iso;
      // cout << "\n";
      if (candidate[map_elts][candidate_iso]
	  && !inv_map[candidate_iso]) {
	map[map_elts] = candidate_iso;
	inv_map[candidate_iso] = true;
	++map_elts;
        if (map_elts < BITS) {
          // even after adding this element, the map isn't complete
          candidate_iso = 0;
        } else {
          // cout << "Checking assertions\n";
          assert(map_elts == BITS);
	  assert(valid_map(map));
	  assert(valid_inv_map(inv_map, map));
          // every position is mapped to
          for (int i=0; i<BITS; ++i) {
            assert(inv_map[i]);
          }
          // cout << "Checking the isomorphism\n";
	  if (weak
	      ? graph -> isomorphic_weak(*(other.graph), map)
	      : (graph -> isomorphic_strong(*(other.graph), map)
		 && induced_transitions_isomorphic(other, map))) {
	    return true;
	  }
          // cout << "Isomorphism failed\n";
          must_increment = true;
	}
      } else {
        // cout << "Bad candidate " << candidate_iso << "\n";
        must_increment = true;
      }

      // The point of this is to avoid duplicating the code.
      if (must_increment) {
        // cout << "Incrementing\n";
	while ((map_elts == BITS)
	       || (candidate_iso == BITS-1)) {
	  // We can't increment this bit, so pop the stack.
	  --map_elts;
          // cout << "Popped the stack; map_elts = " << map_elts
          //      << ", candidate = " << map[map_elts] << "\n";
	  if (map_elts < 0)
	    return false;
	  // If we aren't completely finished...
	  if (map_elts >= 0) {
	    candidate_iso = map[map_elts];
	    inv_map[candidate_iso] = false;
	  }
	}
	++candidate_iso;
      }
    }
    // Shouldn't ever get here.
    assert(false);
    return false;
  }
    
  // I could imagine also returning the offset.
  bool induced_transitions_isomorphic(Gray_Code & other, bit_pos map[BITS]) {
    // Confirm that the transitions are isomorphic.

    // For debugging only.
    // cout << "Returning true for induced_transitions_isomorphic\n";
    // return true;

    if (num_trans != other.num_trans)
      return false;

    // Do I need to try them in both the forward and backward directions?
    // I think the forward direction only is sufficient.
    for (int offset=0; offset<num_trans; ++offset) {
      bool good = true;
      for (int i=0; i<num_trans; i++) {
	if (map[transition_bit[i]] != other.transition_bit[(i+offset) % num_trans]) {
	  good = false;
	  break;
	}
      }
      if (good)
	return true;
    }
    return false;
  }


  // private
  static bool valid_map(bit_pos map[BITS], int map_elts = BITS) {
    {
      bit_pos numeric_inv_map[BITS];
      for (bit_pos i=0; i<BITS; ++i)
	numeric_inv_map[i] = BITS; // initialize with an illegal value
      for (bit_pos i=0; i<BITS; ++i)
	numeric_inv_map[map[i]] = i;
      int num_good = 0;
      for (bit_pos i=0; i<BITS; ++i) {
	assert(numeric_inv_map[i] <= BITS);
	if (numeric_inv_map[i] < BITS)
	  num_good++;
      }
      assert(num_good == map_elts);
    }
    return true;
  }

  static bool valid_inv_map(bool inv_map[BITS], bit_pos map[BITS], int map_elts = BITS) {
    bool computed_inv_map[BITS];
    for (int i=0; i<BITS; i++) computed_inv_map[i] = false;
    for (int i=0; i<map_elts; i++) computed_inv_map[map[i]] = true;
    for (int i=0; i<BITS; i++) assert(inv_map[i] == computed_inv_map[i]);
    return true;
  }


  int girth() {
    if (! (maintain_induced || graph_constraint == obey_input_graph)) {
      cerr << "induced[][] not valid!\n";
      abort();
    }

    for (int i=0; i<BITS; i++)
      if (induced[i][i] != 0)
        return 1;
    for (int i=0; i<BITS; i++)
      for (int j=0; i<BITS; j++)
        if ((induced[i][j] != 0)
            && (induced[j][i] != 0))
          return 2;
    for (int i=0; i<BITS; i++)
      for (int j=0; i<BITS; j++)
        for (int k=0; i<BITS; k++)
          if ((induced[i][j] != 0)
              && (induced[j][k] != 0)
              && (induced[k][i] != 0))
            return 3;
    for (int i=0; i<BITS; i++)
      for (int j=0; i<BITS; j++)
        for (int k=0; i<BITS; k++)
          for (int l=0; i<BITS; l++)
            if ((induced[i][j] != 0)
                && (induced[j][k] != 0)
                && (induced[k][l] != 0)
                && (induced[l][i] != 0))
              return 4;
    cerr << "Big girth!\n";
    abort();
  }
    

  ///
  /// Printing
  ///

  // Print the least-significant-bit first -- in the reverse of the
  // internal representation.
  inline void print_binary(state st) {
    for (int i=0; i<BITS; ++i) {
      cout << ((st & (1 << i)) ? "1" : "0");
    }
  }

  // This is so wordy that it is rarely the right thing to do.
  void print_gray_code() {
    cout << "Gray code of length " << num_trans+1 << " (" << num_trans << " bit changes):\n";
    cout << "state  0: ";
    print_binary(0);
    cout << "\n";
    for (int i=0; i<num_trans; ++i) {
      cout << "                 bit: " << transition_bit[i] << "\n";
      cout << "state ";
      if (i+1 < 10) cout << " ";
      cout << i+1 << ": ";
      print_binary(transition_state[i]);
      cout << "\n";
    }
    cout.flush();
  }

  void print_gray_code_short() {
    cout << "Gray code of length " << num_trans+1 << " (" << num_trans << " bit changes):\n";
    cout << "bits:";
    for (int i=0; i<num_trans; ++i) {
      cout << " " << transition_bit[i];
    }
    cout << "\n";
    cout.flush();
  }

  static void print_square_boolean_array(int a[BITS][BITS]) {
    cout << "   ";		// leading space on row with column headers
    for (int j=0; j<BITS; ++j) {
      cout << j;
    }
    cout << "\n";
    for (int i=0; i<BITS; ++i) {
      cout << i << "  ";
      for (int j=0; j<BITS; ++j) {
	cout << ((a[i][j] == 0) ? " " : "*");
      }
      cout << "\n";
    }
  }

  static void read_square_boolean_array(char * filename, int a[BITS][BITS]) {
    ifstream induced_file(filename);
    read_square_boolean_array(induced_file, a);
    string s;
    induced_file >> s;
    assert(induced_file.eof());
  }

  static void read_square_boolean_array(istream & is, int a[BITS][BITS]) {
    // In this function we don't use assert because the edge file format
    // checks are not merely internal consistency checks.

    char c;
    char c1, c2, c3;
    is.get(c1);
    is.get(c2);
    is.get(c3);
    if ((c1 != ' ') || (c2 != ' ') || (c3 != ' ')) {
      cerr << "First three characters of edge file should be spaces, saw \""
           << c1 << c2 << c3 << "\"\n";
      abort();
    }
    for (int j=0; j<BITS; ++j) {
      is.get(c);
      if (c != j+'0') {
        cerr << "Character " << j+4 << " of edge file should be '" << j
             << "', saw '" << c << "'\n";
        abort();
      }
    }
    string s;
    getline(is, s);
    if (s != "") {
      cerr << "extra characters at end of edge file line 1: \"" << s << "\"\n";
      abort();
    }

    for (int i=0; i<BITS; ++i) {
      is.get(c1);
      is.get(c2);
      is.get(c3);
      if ((c1 != i+'0') || (c2 != ' ') || (c3 != ' ')) {
        cerr << "First three characters of edge file line " << i+2
             << " should be \"" << i << "  \", saw \""
             << c1 << c2 << c3 << "\"\n";
        abort();
      }
      for (int j=0; j<BITS; ++j) {
	is.get(c);
	if (!((c == ' ') || (c == '*'))) {
          cerr << "Edge[" << i << "][" << j << "] should"
               << " be represented by ' ' or '*', saw '" << c << "'\n";
          abort();
        }
	a[i][j] = (c == '*');
      }
      // Not:
      //  is.get(c);
      //  assert(c == '\n');
      // because that assumes a particular end-of-line convention.
      string s;
      getline(is, s);
      // Not assert:  this isn't merely an internal consistency check.
      if (s != "") {
        cerr << "extra characters at end of edge file line " << i+2 << ": \"" << s << "\"\n";
        abort();
      }
    }
  }

  void print_induced() {
    set_graph();
    graph->print_induced();
  }

  void print_map(bit_pos map[BITS]) {
    cout << "Isomorphism:\n";
    for (bit_pos i=0; i<BITS; ++i)
      cout << " " << i;
    cout << "\n";
    for (bit_pos i=0; i<BITS; ++i)
      cout << " " << map[i];
    cout << "\n";
  }

};

// shouldn't this be built in??
int str2int(char *s) {
  int result = 0;
  while (*s != '\0') {
    if (! isdigit(*s)) {
      cerr << "Nondigit " << *s << "\n";
      abort();
    }
    result = result * 10 + (int)(*s - '0');
    s++;
  }
  return result;
}
